function [classPrior,GMM,trainingClassificationRate,testClassificationRate,trainingCLL,trainingLM,trueTrainingLM,sampleMargin] = maxMarginGMM(trainX,trainC,nIter,initGMM,initClassPrior,lambda,testX,testC,eta,updatePrior,factor,useMex)
%[classPrior,GMM,trainingClassificationRate,testClassificationRate,training
%CLL,trainingLM] = maxMarginGMM(trainX,trainC,nIter,initGMM,initClassPrior,lambda,testX,testC,eta,useMex)
%
%Train discriminative GMM using the max-margin criterion.
%
%INPUT:
%trainX:                                    matrix: nTrain x nDim; Holds
%                                           all training samples.
%trainC:                                    vector: nTrain x 1; Holds all
%                                           training labels.
%nIter:                                     scalar; Number of iterations.
%initGMM:                                   cell array: nClass x 1; GMM
%                                           parameters for initialization.
%initClassPrior:                            vector: nClass x 1; Class prior
%                                           for initialization.
%lambda:                                    scalar; Design parameter.
%testX:                                     matrix: nTest x nDim; Holds
%                                           all test samples for cross validation.
%testC:                                     vector: nTest x 1; Holds all
%                                           test labels.
%eta:                                       scalar; Parameter for
%                                           max-approximation.
%updatePrior:                               bool; If true, prior will be
%                                           updated too.
%factor:                                    scalar; Once all variances are
%                                           positive, D1 will be multiplied by this factor. If factor is set to zero,
%                                           D1 will be doubled until trueTrainingLM increases (or a threshold for D1 is reached)
%useMex:                                    bool; If true, the MATLAB MEX
%                                           implementation for faster evaluation of the GMM will be used.
%
%OUTPUT:
%classPrior:                                vector: nClass x 1; The resulting
%                                           class prior.
%GMM:                                       cell array: nClass x 1; The
%                                           resulting GMM parameters.
%trainingClassificationRate:                vector: 1 x nIter; Percentage
%                                           of correctly classified samples on training set after each iteration.
%testClassificationRate:                    vector: 1 x nIter; Same for
%                                           test set.
%trainingCLL:                               vector: 1 x nIter; Conditional
%                                           log-likelihood on training set after each iteration.
%trainingLM:                                vector: 1 x nIter; approximated (soft-hinge) cost function after each iteration.                                          
%trueTrainingLM:                            vector; 1 x nIter. Exact cost function.
%sampleMargin:                              matrix: nTrain x nIter; Margin d of each training sample after each iteration.


%author: Michael Wohlmayr, 19.02.2010
%v.0.1

%% PARAMETERS

UPPERLIMIT_D1 = 2^300; %If factor==0, we will double D1 no more than this value

%% Check input
[nTrain,nDim] = size(trainX);


%%
%We assume that training class labels in trainC are within the range
%[1,max(trainC)].
%For some classes, it might happen that no training data is available. For
%this purpose, we create an internal class assignment which omits those
%classes without training data.
nExternalClass = max(trainC);
internalClass = zeros(nTrain,1);
newClassId = 1;
nInternalClass = length(unique(trainC));
externalClass = zeros(1,nInternalClass);
for kExternalClass = 1:nExternalClass
    classSampleIndex = (trainC==kExternalClass);
    if(any(classSampleIndex))
        internalClass(classSampleIndex) = newClassId; %assign every training sample an internal class ID
        externalClass(newClassId) = kExternalClass;
        newClassId = newClassId + 1;
    end
end

%%
%Discard samples from test data which belong to class that we omitted in
%the previous step. Assign internal class labels.
newC = zeros(1,size(testX,1));
for kClass = 1:nInternalClass
    newC(testC==externalClass(kClass)) = kClass;
end
testX = testX(newC~=0,:);
testC = newC(newC~=0);
nTest = size(testX,1);

%%
%Convert GMM parameters
mu = cell(1,nInternalClass); %mean
sigma = cell(1,nInternalClass); %diagonal covariance
w = cell(1,nInternalClass); %component priors
rho = zeros(1,nInternalClass); %class prior
nComp = zeros(1,nInternalClass); %number of components of each GMM

for kClass = 1:nInternalClass
    currentExternalClass = externalClass(kClass);
    nComp(kClass) = length(initGMM{currentExternalClass});
    mu{kClass}=zeros(nDim,nComp(kClass));
    sigma{kClass}=zeros(nDim,nComp(kClass));
    w{kClass}=zeros(1,nComp(kClass));
    rho(kClass) = initClassPrior(currentExternalClass);
    
    for kComp = 1:nComp(kClass)
        w{kClass}(1,kComp) = initGMM{currentExternalClass}{kComp}.prior;
        mu{kClass}(:,kComp) = initGMM{currentExternalClass}{kComp}.mean;
        sigma{kClass}(:,kComp) = diag(initGMM{currentExternalClass}{kComp}.cov);
    end
end

%% Pre-compute some quantities
u = zeros(nTrain,nInternalClass);
for kClass = 1:nInternalClass
    u(:,kClass) = internalClass==kClass;
end


%%
%perform one E-step before iteration starts
[s,X1,X2,A1,A2,actTrainingLM,actTrueTrainingLM,approxPartialRho,approxPartialAlpha,logPxc,logPxcTest,margin] = EStep(mu,sigma,w,rho,trainX,testX,u,lambda,eta,nInternalClass,internalClass,nComp,useMex);


%%allocate memory for diagnostics
trainingClassificationRate = zeros(1,nIter);
testClassificationRate = zeros(1,nIter);
trainingCLL = zeros(1,nIter);
trainingLM = zeros(1,nIter);
trueTrainingLM = zeros(1,nIter);
sampleMargin = zeros(nTrain,nIter);


for kIter = 1:nIter
    
    fprintf('Iteration %d\n',kIter);
    
    
    if(kIter==1)
        previousTrueTrainingLM = -Inf;
    else
        previousTrueTrainingLM = trueTrainingLM(kIter-1); %We will use this variable to check if trueTrainingLM increases
    end
    
    %% M-Step
    %Compute new parameters.
    %Constant D1 is increased until all variances in sigma are positive.
    %Further, if factor==0, we will increase D1 until cost function is
    %nondecreasing or an upper limit on D1 is reached.
    %If factor>0, only that factor will be multiplied with D1.
    
    
    %prepare some quantities for fast scanning of best D1
    partMuNum = cell(1,nInternalClass);
    partSigmaNum = cell(1,nInternalClass);
    Den = cell(1,nInternalClass);
    for kClass = 1:nInternalClass
        
        partMuNum{kClass} = zeros(nDim,nComp(kClass));
        partSigmaNum{kClass} = zeros(nDim,nComp(kClass));
        Den{kClass} = zeros(1,nComp(kClass));
        
        for kComp = 1:nComp(kClass)
            
            a1TimesS = s{kClass}(X1,kComp).*A1;
            a2TimesS = s{kClass}(X2,kComp).*A2;
            
            %mu
            partMuNum{kClass}(:,kComp) = sum(a1TimesS(:,ones(1,nDim)).*trainX(X1,:)).' + ...
                sum(a2TimesS(:,ones(1,nDim)).*trainX(X2,:)).';
            
            %sigma
            partSigmaNum{kClass}(:,kComp) = sum(a1TimesS(:,ones(1,nDim)).*trainX(X1,:).^2).' + ...
                sum(a2TimesS(:,ones(1,nDim)).*trainX(X2,:).^2).';
            
            %the denomiator is the same for both mu and sigma
            Den{kClass}(kComp) = sum(a1TimesS) + sum(a2TimesS);
            
        end
    end
    
    %Compute mu and sigma. As soon as a sigma gets negative, restart
    %computations with double D1.
    D1 = 1;
    muNew = cell(1,nInternalClass);
    sigmaNew = cell(1,nInternalClass);
    negativeSigmaFound = 1;
    breakLoop = 0;
    while(negativeSigmaFound)
        
        negativeSigmaFound = 0;
        for kClass = 1:nInternalClass
            for kComp = 1:nComp(kClass)
                
                muOld = mu{kClass}(:,kComp);
                muOldSquare = muOld.^2;
                
                %mu
                muNew{kClass}(:,kComp) = (partMuNum{kClass}(:,kComp) + D1*muOld)/(Den{kClass}(kComp) + D1);
                
                %sigma
                sigmaNew{kClass}(:,kComp) = (partSigmaNum{kClass}(:,kComp) + D1*(sigma{kClass}(:,kComp)+muOldSquare))/(Den{kClass}(kComp) + D1) - muNew{kClass}(:,kComp).^2;
                
                %check if any variance is negative
                if(any(sigmaNew{kClass}(:,kComp)<0))
                    %restart the double loop with doubled D1.
                    D1 = D1 * 2;
                    negativeSigmaFound = 1;
                    breakLoop = 1;
                    break;
                end
                
            end
            
            if(breakLoop)
                breakLoop = 0;
                break;
            end
        end
        
    end
    
    wNew = cell(1,nInternalClass);
    
    %If factor==0, double D1 until cost increases.
    %This while-iteration involves the E-Step for the (possibly) new parameters
    %and needs to be stepped through at least once.
    costIncreased = 0;
    while(~costIncreased)
        if(factor==0)
            D1 = D1 * 2;
            fprintf('Cost not increased;  doubling D1 to %d\n',D1);
        else
            D1 = D1*factor;
            fprintf('D1 <- D1*factor = %d\n',D1);
        end
        
        
        
        %new mu and sigma
        for kClass = 1:nInternalClass
            for kComp = 1:nComp(kClass)
                
                muOld = mu{kClass}(:,kComp);
                muOldSquare = muOld.^2;
                
                %mu
                muNew{kClass}(:,kComp) = (partMuNum{kClass}(:,kComp) + D1*muOld)/(Den{kClass}(kComp) + D1);
                
                %sigma
                sigmaNew{kClass}(:,kComp) = (partSigmaNum{kClass}(:,kComp) + D1*(sigma{kClass}(:,kComp)+muOldSquare))/(Den{kClass}(kComp) + D1) - muNew{kClass}(:,kComp).^2;
                
            end
        end
        
        %new rho
        if(updatePrior)
            rhoNew = rho.*(approxPartialRho + D1);
            rhoNew = rhoNew./sum(rhoNew);
        end
        
        %new w
        for kClass = 1:nInternalClass
            if(nComp(kClass)==1)
                wNew{kClass} = 1;
            else
                wNewTmp = w{kClass}.*(approxPartialAlpha{kClass} + D1);
                wNew{kClass} = wNewTmp/sum(wNewTmp);
            end
        end
        
        %E-Step with new parameters
        [s,X1,X2,A1,A2,actTrainingLM,actTrueTrainingLM,approxPartialRho,approxPartialAlpha,logPxc,logPxcTest,margin] = EStep(mu,sigma,w,rho,trainX,testX,u,lambda,eta,nInternalClass,internalClass,nComp,useMex);
        
        
        
        if(actTrueTrainingLM>previousTrueTrainingLM)
            %fine, now we can go on with this value for D1
            costIncreased = 1;
        end
        if(D1 > UPPERLIMIT_D1)
            %we give up...
            costIncreased = 1;
            fprintf('Iteration %d: D1 reached upper limit\n',kIter);
        end
        if(factor ~= 0)
            costIncreased = 1; %Factor>0: We go through this just once, as we don't care about the cost.
        end
        
    end
    %% end M-Step
    
    
    mu = muNew;
    sigma = sigmaNew;
    w = wNew;
    if(updatePrior)
        rho = rhoNew;
    end
    
    
    %% Classification rate
    [ignore,currentClassPrediction] = max(logPxc,[],2); %training set
    trainingClassificationRate(kIter) = sum(currentClassPrediction==internalClass)/nTrain*100;
    fprintf('Classification Rate on Training Set: %f\n',trainingClassificationRate(kIter));
    [ignore,currentClassPrediction] = max(logPxcTest,[],2); %test set
    testClassificationRate(kIter) = sum(currentClassPrediction==testC.')/nTest*100;
    fprintf('Classification Rate on Test Set: %f\n',testClassificationRate(kIter));
    clear currentClassPrediction;
    clear ignore;
    
    %% Conditional Log-Likelihood
    tmp = logPxc(logical(u)) - logsumexp(logPxc,2);
    trainingCLL(kIter) = sum(tmp);
    clear logPxc tmp;
    
    %% margin cost function
    trainingLM(kIter) = actTrainingLM;
    trueTrainingLM(kIter) = actTrueTrainingLM;
    
    sampleMargin(:,kIter) = margin;
    
    
end





classPrior = rho;
GMM = initGMM;
%write modified parameters to GMM
for kClass = 1:nInternalClass
    for kComp = 1:nComp(kClass)
        GMM{externalClass(kClass)}{kComp}.prior = w{kClass}(kComp);
        GMM{externalClass(kClass)}{kComp}.mean = mu{kClass}(:,kComp);
        GMM{externalClass(kClass)}{kComp}.cov = diag(sigma{kClass}(:,kComp));
    end
end


function [pX,pXTest,gamma] = evalGMM(mu,sigma,w,trainX,testX,nComp,nInternalClass)
%Given a set of |C| GMMs, compute log p(X|C) on training and test data, as well
%as p(k|X,C) for the training data (where k is the component index of the
%C-th GMM).
%
%INPUT:
%mu:                cell array; mu{c}(:,k) has -kth mean of c-th GMM.
%sigma:             cell array; sigma{c}(:,k) has -kth diagonal covariance of c-th GMM.
%w:                 cell array; w{c}(k) has k-th weight of c-th GMM.
%trainX:            matrix; trainX(n,:) has n-th training sample.
%testX:             matrix; trainX(n,:) has n-th test sample.
%nComp:             vector; nComp(c) has number of components of c-th GMM.
%nInternalClass     scalar; number of GMMs
%
%OUTPUT:
%pX:                matrix; pX(n,c) gives log p(x^n|C==c) for training data.
%pXTest:            matrix; pX(n,c) gives log p(x^n|C==c) for test data.
%gamma:             cell array; gamma{c}(n,k) gives component posterior
%                   p(k|x^n,C==c).
%
%evalDataMEX is the equivalent implementation of this function using MEX.

%author: Michael Wohlmayr, 19.02.2010

[nTrain,nDim] = size(trainX);
nTest = size(testX,1);
const1 = nDim*0.5*log(2*pi); %for computation of normal distribution
gamma = cell(1,nInternalClass);
pX = zeros(nTrain,nInternalClass); %p(x^n|\theta_c) on training set
pXTest = zeros(nTest,nInternalClass); %p(x^n|\theta_c) on test set
for kClass = 1:nInternalClass
    
    gamma{kClass} = zeros(nTrain,nComp(kClass));
    logW = log(w{kClass});
    
    actMuMatrix = mu{kClass}.';
    actStdMatrix = sqrt(sigma{kClass}).';
    logPXallComp = zeros(nTrain,nComp(kClass));
    logPXTestAllComp = zeros(nTest,nComp(kClass));
    for kComp = 1:nComp(kClass)
        
        actMu = actMuMatrix(kComp,:);
        actStd = actStdMatrix(kComp,:);
        logActStd = log(actStd);
        
        %probability of trainX for all GMM components
        Z = (trainX-actMu(ones(1,nTrain),:))./actStd(ones(1,nTrain),:); %standardize: Z = (X-mu)/sigma
        Z = Z.^2;
        logPXallComp(:,kComp) = -const1-sum(0.5*Z+logActStd(ones(1,nTrain),:),2);

        
        %probability of testX for all GMM components
        Z = (testX-actMu(ones(1,nTest),:))./actStd(ones(1,nTest),:); %standardize: Z = (X-mu)/sigma
        Z = Z.^2;
        logPXTestAllComp(:,kComp) = -const1-sum(0.5*Z+logActStd(ones(1,nTest),:),2);
        
    end
    
    %gamma and p(x^n|\theta_c) on training set
    if(nComp(kClass)==1)
        gamma{kClass} = ones(nTrain,1);
        pX(:,kClass) = logPXallComp + logW(ones(nTrain,1));
    else
        actGamma = logPXallComp + logW(ones(nTrain,1),:);
        normGamma = logsumexp(actGamma,2);
        actGamma = actGamma - normGamma(:,ones(1,nComp(kClass)));
        gamma{kClass} = exp(actGamma);
        
        %p(x^n|\theta_c) for training set
        pX(:,kClass) = normGamma;    
    end
    
    
    
    %p(x^n|\theta_c) for test set
    pXTest(:,kClass) = logsumexp(logPXTestAllComp + logW(ones(nTest,1),:),2);
    
end

function [s,X1,X2,A1,A2,trainingLM,trueTrainingLM,approxPartialRho,approxPartialAlpha,logPxc,logPxcTest,margin] = EStep(mu,sigma,w,rho,trainX,testX,u,lambda,eta,nInternalClass,internalClass,nComp,useMex)

nTrain = size(trainX,1);
nTest = size(testX,1);

if(useMex)
    %use fast MEX-implementation
    [pX,gamma,pXTest] = evalGMMMEX(mu,sigma,w,trainX,testX);
    
    %check for infinite values in gamma
    for kClass = 1:nInternalClass
        actGamma = gamma{kClass};
        actGamma(~isfinite(actGamma)) = 0;
        gamma{kClass} = actGamma;
    end
    
else
    %use MATLAB-implementation
    [pX,pXTest,gamma] = evalGMM(mu,sigma,w,trainX,testX,nComp,nInternalClass);
end

logRho = log(rho);

%log p(X,C) on training set
logPxc = pX + logRho(ones(nTrain,1),:);

%log p(X,C) on test set
logPxcTest = pXTest + logRho(ones(nTest,1),:);

%r
rNorm = logsumexp(logPxc + log(double(~u)),2);
r = eta*(logPxc - rNorm(:,ones(1,nInternalClass)));
r = exp(r);

%notUTimesR = (~u).*r;
notUTimesR = r;
notUTimesR(logical(u)) = 0;

%We will need several times the term gamma*(u - ~u*r)
s = cell(1,nInternalClass);
for kClass = 1:nInternalClass
    part = u(:,kClass) - notUTimesR(:,kClass);
    s{kClass} = gamma{kClass}.*part(:,ones(1,nComp(kClass)));
end



%% Margin of training samples
% We use the exact maximum for Denomiator.
pXTrans = pX.';
Num = logRho(internalClass).'+ pXTrans(logical(u).');
Den = max(pX + log(double(~u)) + logRho(ones(nTrain,1),:),[],2); %log(max(.))=max(log(.))
margin = exp(Num - Den); %might get Inf, but that doesn't matter
clear pXTrans;

%get three regions of data depending on margin
X1 = margin<=1.^(1/lambda);
X3 = margin>=2.^(1/lambda);
X2 = ~(X1|X3);

%LMtotal
y = margin.^lambda;
trainingLM = sum(log(y(X1)+0.5)) + sum(log(2-0.5*(y(X2)-2).^2)) + sum(X3)*log(2);
trueTrainingLM = sum(min(lambda*(Num-Den),log(2))); %use log of margin (Num-Den)

%The weight for samples from X1 and X2 (denoted as s^n in pseudo code)
A1 = margin(X1).^lambda;
A1 = (lambda*A1)./(A1 + 0.5);
A2 = margin(X2).^lambda;
A2 = lambda*(2-A2)./(2-0.5*A2);

%approximate \frac{\partial \log D}{\partial \rho_c} using two
%fractions: Num1/Den1 - Num2/Den2
Num1 = sum(A1(:,ones(1,nInternalClass)).*u(X1,:)) + sum(A2(:,ones(1,nInternalClass)).*u(X2,:));
Den1 = sum(Num1);

Num2 = sum(A1(:,ones(1,nInternalClass)).*notUTimesR(X1,:)) + sum(A2(:,ones(1,nInternalClass)).*notUTimesR(X2,:));
Den2 = sum(Num2);

approxPartialRho = Num1/Den1-Num2/Den2;

approxPartialAlpha = cell(1,nInternalClass);
for kClass = 1:nInternalClass
    
    %approximate \frac{\partial \log D}{\partial \alpha_c^m} using two
    %fractions: Num1/Den1 - Num2/Den2
    part1 = u(X1,kClass).*A1;
    part2 = u(X2,kClass).*A2;
    Num1 = sum(gamma{kClass}(X1,:).*part1(:,ones(1,nComp(kClass)))) + sum(gamma{kClass}(X2,:).*part2(:,ones(1,nComp(kClass))));
    Den1 = sum(Num1);
    
    
    part1 = notUTimesR(X1,kClass).*A1;
    part2 = notUTimesR(X2,kClass).*A2;
    Num2 = sum(gamma{kClass}(X1,:).*part1(:,ones(1,nComp(kClass)))) + sum(gamma{kClass}(X2,:).*part2(:,ones(1,nComp(kClass))));
    Den2 = sum(Num2);
    
    approxPartialAlpha{kClass} = Num1/Den1-Num2/Den2;
    
end


